
# This file was generated by 'versioneer.py' (0.22) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2023-03-08T17:19:55+1100",
 "dirty": false,
 "error": null,
 "full-revisionid": "99731ea365df1a59de674ca3b6326727289ee881",
 "version": "0.16.3.post11"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
